from SlatAssembly import *
import SlatType as TASSlats
import xml.etree.ElementTree as ET
import xml.dom.minidom
import copy


MODEL_ABSTRACT = 0
MODEL_KINETIC  = 1

# The following colors aren't in the standard wxPython database so the listed conversions will be made for viewing in SlatSimulator
wxPython_color_mapping = {'aqua': 'AQUAMARINE', 'lime': 'LIME GREEN', 'olive': 'KHAKI'}

class SlatSystem:
    def __init__(self):
        self.Clear()
        self.binding_threshold = 2
        self.model = MODEL_ABSTRACT
        self.b_use_rotation = False
        
    def Clear(self):
        self.binding_threshold = None
        self.PolyominoTypeBase_dict = dict()
        self.PolyominoType_dict = dict()
        self.SeedPolyominoType_dict = dict()
        self.seed = SlatAssembly(self)
        self.assembly = SlatAssembly(self)
        self.concentrations_dict = dict()
        self.concentrations_total = 0
        self.Gmc = None
        self.Gse = None
        self.max_polyomino_binding = 0
        self.stochastic_event_set = None

    def GetPolyominoTypeByName(self, type_name):
        for p_type_name in self.PolyominoType_dict:
            if p_type_name == type_name:
                return self.PolyominoType_dict[p_type_name]
        return None

    def AddPolyominoType(self, pct):
        if pct.name in self.PolyominoType_dict:
            print(f"PolyominoType with name {pct.name} already exists in set")
        else:
            self.PolyominoTypeBase_dict[pct.name] = pct
            rot = pct.Copy()
            rot.base_type = pct
            self.PolyominoType_dict[rot.name] = rot
            if self.b_use_rotation:
                rotPolys = self.GetRotatedPolyominoes(pct)
                #print(f'Rotated polyominoes: {rotPolys}')
                for rp in rotPolys:
                    rp.base_type = pct
                    self.PolyominoType_dict[rp.name] = rp

            if not pct.concentration is None:
                self.concentrations_total = self.concentrations_total + pct.concentration
                if pct.concentration in self.concentrations_dict:
                    self.concentrations_dict[pct.concentration].append(pct.name)
                else:
                    self.concentrations_dict[pct.concentration] = [pct.name]
            max_binding = pct.GetMaxBinding()
            self.max_polyomino_binding = max(self.max_polyomino_binding, max_binding)

    # This gets copies rotated around the z axis in 90 degree increments, counterclockwise
    def GetRotatedPolyominoes(self, pct):
        rot_polys = []
        for angle in [90,180,270]:
            pctR = pct.GetRotatedVersion(angle)
            rot_polys.append(pctR)
        return rot_polys

    def ReadFromFile(self, file_name, b_viewing_only=False):
        self.Clear()
        xml_tree = ET.parse(file_name)
        xml_root = xml_tree.getroot()
        dim_restrictions = None
        min_binding = None
        if not xml_root.tag == 'PolyominoSystem':
            print("Invalid XML PolyominoSystem file")
            return False
        else:
            seed_polyominoes = []
            assembly_polyominoes = []
            dim_restrictions = (None, None, None)
            for child in xml_root:
                if child.tag == 'bindingThreshold':
                    self.binding_threshold = int(child.text)
                elif child.tag == 'Gmc':
                    self.Gmc = float(child.text)
                elif child.tag == 'Gse':
                    self.Gse = float(child.text)
                elif child.tag == 'rotation':
                    if child.text in ['True', 'true', 't', 'T']:
                        self.b_use_rotation = True
                    else:
                        self.b_use_rotation = False
                elif child.tag == 'dim_restrictions':
                    dim_restrictions = eval(child.text)
                elif child.tag == 'min_binding':
                    min_binding = int(child.text)
                elif child.tag == 'PolyominoTypes':
                    for elt_type in child:
                        pct = TASSlats.SlatType(name=None, xml_node=elt_type, dim_restrictions = dim_restrictions)
                        if pct.b_seed_only == True:
                            self.SeedPolyominoType_dict[pct.name] = pct
                        else:
                            self.AddPolyominoType(pct)
                    print(f'Loaded {len(self.PolyominoType_dict)} polyomino types')
                elif child.tag == 'seed':
                    for child2 in child:
                        if child2.tag == 'Polyominoes':
                            for elt_polyomino in child2:
                                p = TASSlats.Slat(slatType=None, xml_node=elt_polyomino, slatTypes=[self.PolyominoType_dict,self.SeedPolyominoType_dict])
                                seed_polyominoes.append(p)
                    print(f'Loaded {len(seed_polyominoes)} polyominoes in the seed')
                elif child.tag == 'assembly':
                    for child2 in child:
                        if child2.tag == 'Polyominoes':
                            for elt_polyomino in child2:
                                p = Slat(slatType=None, xml_node=elt_polyomino, slatTypes=[self.PolyominoType_dict,self.SeedPolyominoType_dict])
                                assembly_polyominoes.append(p)
                    print(f'Loaded {len(assembly_polyominoes)} polyominoes in the assembly')
        
        # If Gmc and Gse were set, we're using the kinetic model
        if (not self.Gmc is None) and (not self.Gse is None):
            self.model = MODEL_KINETIC
            print("Simulation model set to 'kinetic'")
        else:
            self.model = MODEL_ABSTRACT
            print("Simulation model set to 'abstract'")
            
        if not dim_restrictions is None:
            print(f"Using dimension restrictions {dim_restrictions}")
            
        self.assembly.polyominoSystem = self
        self.seed.polyominoSystem = self

        if not min_binding is None:
            self.assembly.min_binding = min_binding
            self.seed.min_binding = min_binding

        if len(seed_polyominoes) > 0:
            #print(f'Adding {len(seed_polyominoes)} polyominoes in the seed...')
            for p in seed_polyominoes:
                #if b_viewing_only:
                self.seed.AddPolyominoForViewingOnly(p)
                #else:
                    #self.seed.AddPolyomino(p)
            print(f'Added {len(seed_polyominoes)} polyominoes from the seed assembly')
                    
        if (len(assembly_polyominoes) > 0) or (len(seed_polyominoes) > 0):
            print(f'Adding {len(assembly_polyominoes)} non-seed polyominoes in the assembly...')
            for polyomino_list in [seed_polyominoes, assembly_polyominoes]:
                for p in polyomino_list:
                    #if b_viewing_only:
                    self.assembly.AddPolyominoForViewingOnly(p)
                    #else:
                        #self.assembly.AddPolyomino(p)
            print(f'Added {len(assembly_polyominoes)} polyominoes to the assembly')


        self.assembly.setDimensions()

        # Validation steps:
        # - if given types and seed, verify that each entry in the seed is of one of the types
        # - if given types and assembly, verify that each entry in the assembly is of one of the types
        for loc in self.seed.locationMap:
            polyomino = self.seed.locationMap[loc]
            if (not polyomino.slatType.name in self.PolyominoType_dict) and (not polyomino.slatType.name in self.SeedPolyominoType_dict):
                print(f"ERROR: seed contains Polyomino of type {polyomino.polyominoType.name} which is not in the PolyominoType set {self.PolyominoType_dict.keys()}")
                return False
        for loc in self.assembly.locationMap:
            polyomino = self.assembly.locationMap[loc]
            if (not polyomino.slatType.name in self.PolyominoType_dict) and (not polyomino.slatType.name in self.SeedPolyominoType_dict):
                print(f"ERROR: assembly contains Polyomino of type {polyomino.polyominoType.name} which is not in the PolyominoType set {self.PolyominoType_dict.keys()}")
                return False
        return True   
                        
    def WriteToFile(self, file_name, color_mapping=wxPython_color_mapping):
        xml_root = ET.Element('PolyominoSystem')
        if not self.binding_threshold is None:
            elt_bt = ET.SubElement(xml_root, 'bindingThreshold')
            elt_bt.text = str(self.binding_threshold)
        if not self.Gmc is None:
            elt_gmc = ET.SubElement(xml_root, 'Gmc')
            elt_gmc.text = str(self.Gmc)
        if not self.Gse is None:
            elt_gse = ET.SubElement(xml_root, 'Gse')
            elt_gse.text = str(self.Gse)
        if not self.b_use_rotation is None:
            elt_gse = ET.SubElement(xml_root, 'rotation')
            elt_gse.text = str(self.b_use_rotation)
        if not self.PolyominoType_dict is None:
            elt_pct_set = ET.SubElement(xml_root, 'PolyominoTypes')
            for pct in self.PolyominoType_dict:
                self.PolyominoType_dict[pct].toXML(elt_pct_set, color_mapping)
        if not self.seed is None:
            elt_seed = ET.SubElement(xml_root, 'seed')
            elt_polyominoes = ET.SubElement(elt_seed, 'Polyominoes')
            polyominoes = self.seed.GetPolyominoes()
            count = 0
            for polyomino in polyominoes:
                polyomino.toXML(elt_polyominoes)
                count += 1
            print(f'Output {count} seed slats')
        if not self.assembly is None:
            elt_assembly = ET.SubElement(xml_root, 'assembly')
            elt_polyominoes = ET.SubElement(elt_assembly, 'Polyominoes')
            seed_polyominoes = self.seed.GetPolyominoes()
            polyominoes = self.assembly.GetPolyominoes()
            count = 0
            for polyomino in polyominoes:
                if not polyomino in seed_polyominoes:
                    polyomino.toXML(elt_polyominoes)
                    count += 1
            print(f'Output {count} assembly slats')

        elt_dim_restrictions = ET.SubElement(xml_root, 'dim_restrictions')
        elt_dim_restrictions.text = '(None,None,[0,2])'

        xml_string = ET.tostring(xml_root)
        xml_dom = xml.dom.minidom.parseString(xml_string)
        
        with open(file_name, "w") as f: 
            f.write(xml_dom.toprettyxml())
        return True
