from SlatType import Slat as TASSlat
from SlatSystem import *
from slat_generator import *
import sys
import getopt


COLORING_BITS = 0
COLORING_MOTIF = 1


def DisplayHelp():
    print("\nThis script takes as input a scale factor and outputs a set of aTAM tiles and/or a set")
    print("of slats for that scale factor which self-assembly the Sierpinski triangle pattern.")
    print("\nUsage:")
    print("-h, --help\t\t\tDisplays this help message.")
    print("-n <name>, --name=<name>\t[Optional] Base name of the output files. If not specified,")
    print("\t\t\t\t 'slat-Sierpinski-triangle-scale<scale>' is used (with the scale factor substituted for <scale>.")
    print("-o <dir>, --out=<dir>\t\t[Optional] Directory for output files. If not specified,")
    print("\t\t\t\t the current directory will be used.")
    print("-l <scale>, --scale=<scale>\t[Optional] Size of slat macrotiles. If not specified, defaults to 4.")
    print("-m <motif>, --motif=<motif>\t[Optional] If specified, the pattern of slats in each macrotile will be that of")
    print("\t\t\t\t the specified motif, e.g. 'abbaba'.")
    print("-x, --xml\t\t\t[Optional] If specified, slats will be output to a XML file.")
    print("-t, --tiles\t\t\t[Optional] If specified, TDS and TDP files will be output.")
    print("-s, --svg\t\t\t[Optional] If specified, slats will be output to an SVG file.")
    print("-p, --pdf\t\t\t[Optional] If specified, slats will be output to a PDF file.")
    print("-f <T/F>, --flip=<T/F>\t\t[Optional] If this option is set to 'F', versions of slat types rotated 180 degrees")
    print("\t\t\t\t will not be included. Otherwise, they will.")
    print("-c <b/m>, --color=<b/m>\t\t[Optional] If this option is set to 'm', slats will be colored based on the motif")
    print("\t\t\t\t pattern. Otherwise, they will be colored to match the bit pattern.")
    print("")
    exit()


########################################
## Definitions of gadget templates and functions to generate the dictionaries for their instantiations
##   Each gadget template is a list of tile types, where each tile type is a dict mapping tile sides to glue functions
##  (see glue_functions in slat_generator.py), None, or a number for a unique tau-strength glue for connecting tiles of that gadget.

Seed_gadget = [{'coords':(-1,0), N:outtau1, E:(1,intau), S:None, W:None},
               {'coords':(0,0), N:None, E:(2,outtau), S:None, W:(1,outtau)},
               {'coords':(1,0), N:outtau2, E:None, S:None, W:(2,intau)}]
def DictGenerator_Seed(var_dict=None):
    glue_dict = {outtau1:'left', outtau2:'right',
                 'name': f'SEED', 'label': f'SEED', 'color': 'red'}
    return [glue_dict]


Left_gadget = [{'coords':(-1,0), N:outtau1, E:(1,intau), S:None, W:None},
               {'coords':(0,0), N:None, E:out1, S:intau, W:(1,outtau)}]
def DictGenerator_Left(var_dict=None):
    glue_dict = {outtau1:'left', out1:1, intau:'left',
                 'name': f'LEFT', 'label': f'LEFT', 'color': 'red'}
    return [glue_dict]


Right_gadget = [{'coords':(0,0), N:None, E:(1,outtau), S:intau, W:out1},
                {'coords':(1,0), N:outtau1, E:None, S:None, W:(1,intau)}]
def DictGenerator_Right(var_dict=None):
    glue_dict = {outtau1:'right', out1:1, intau:'right',
                 'name': f'RIGHT', 'label': f'RIGHT', 'color': 'red'}
    return [glue_dict]


Interior_gadget = [{'coords':(0,0), N:(1,outtau), E:in1, S:None, W:in2},
                   {'coords':(0,1), N:None, E:out1, S:(1,intau), W:out1}]
def DictGenerator_Interior(var_dict=None):
    dict_list = []
    for input1 in [0,1]:
        for input2 in [0,1]:
            output_bit = input1^input2
            color = 'white' if output_bit == 0 else 'red'
            glue_dict = {in1:f'{input1}', in2:f'{input2}', out1:f'{output_bit}',
                         'name': f'{input1},{input2}', 'label': f'{output_bit}', 'color': color}
            dict_list.append(glue_dict)
            #print(f'len: {len(glue_dict)}, glue_dict: {glue_dict}')
    return dict_list


def GetTilesAndSlats(scale_factor, motif=None):

    ########################################
    ## Generate tiles and slats for the boundary and interior gadgets
    gadget_tiles_dict = dict()
    gadget_slat_groups = []
    
    gadget_definitions = [(Left_gadget, DictGenerator_Left), (Right_gadget, DictGenerator_Right), (Interior_gadget, DictGenerator_Interior)]

    for gadget_definition in gadget_definitions:
        (tile_dict, slat_groups) = GetTilesAndSlatsHelper(gadget_definition[0], gadget_definition[1], scale_factor, motif=motif)
        combineTileDicts(tile_dict, gadget_tiles_dict)
        gadget_slat_groups += slat_groups

    ########################################
    ## Generate tiles and slats for the seed gadget
    seed_tiles_dict = dict()
    seed_slat_groups = []
    seed_name = 'SEED2'

    gadget_definitions = [(Seed_gadget, DictGenerator_Seed)]

    for gadget_definition in gadget_definitions:
        (tile_dict, slat_groups) = GetTilesAndSlatsHelper(gadget_definition[0], gadget_definition[1], scale_factor, motif=motif)
        combineTileDicts(tile_dict, seed_tiles_dict)
        seed_slat_groups += slat_groups

    # Leave out the seed since seed types are declared to only be used in the seed and we don't want to remove their
    # matching pairs if they exist
    #MergeMatchingSlats([gadget_slat_groups, seed_slat_groups])

    MergeMatchingSlats([gadget_slat_groups])

    combineTileDicts(seed_tiles_dict, gadget_tiles_dict)
    print(f'Tile set contains {len(gadget_tiles_dict)} tile types. ({len(seed_tiles_dict)} are seed tile types, {len(gadget_tiles_dict)-len(seed_tiles_dict)} are regular tile types.)')

    slat_count = 0
    for slat_group in gadget_slat_groups:
        slat_count = slat_count + len(slat_group.horizontal_slat_group.slats)
        slat_count = slat_count + len(slat_group.vertical_slat_group.slats)

    seed_slat_count = 0
    for slat_group in seed_slat_groups:
        seed_slat_count = seed_slat_count + len(slat_group.horizontal_slat_group.slats)
        seed_slat_count = seed_slat_count + len(slat_group.vertical_slat_group.slats)

        # Ensure that slats in the seed-specific gadgets are marked as only being allowed in the seed assembly
        for slat in slat_group.horizontal_slat_group.slats:
            slat.b_seed_only = True
        for slat in slat_group.vertical_slat_group.slats:
            slat.b_seed_only = True
 
    #print(f'Slat set contains {slat_count+seed_slat_count} slat types. ({seed_slat_count} are seed slat types, {slat_count} are regular slat types.)')
    return gadget_tiles_dict, seed_name, gadget_slat_groups, seed_slat_groups

def SetSierpinskiSlatColor(tas_slat_type=None, lib_slat_type=None):
    color = None
    slat_type = tas_slat_type if lib_slat_type is None else lib_slat_type
    if slat_type.name.startswith('SEED'):
        color = 'green'
    else:
        if slat_type.name.startswith('LEFT') or slat_type.name.startswith(
               'RIGHT') or slat_type.name.startswith('1,0') or slat_type.name.startswith('0,1'):
            color = 'blue'
        else:
            color = 'yellow'

    if not tas_slat_type is None:
        if tas_slat_type.IsHorizontal():
            color = None
        tas_slat_type.color = color
        tas_slat_type.colors = dict()

    if not lib_slat_type is None:
        if lib_slat_type.b_horizontal:
            color = None
        lib_slat_type.color = color
        for i in range(len(lib_slat_type.io_locs)):
            lib_slat_type.io_locs[i] = (lib_slat_type.io_locs[i][0], lib_slat_type.io_locs[i][1], color, lib_slat_type.io_locs[i][3])


if __name__ == "__main__":
    try:
        # starts at the second element of argv since the first one is the script name
        # extraparams are extra arguments passed after all option/keywords are assigned
        # opts is a list containing the pair "option"/"value"
        opts, extraparams = getopt.getopt(sys.argv[1:], "hn:o:l:xtspf:m:c:",
                            ["help", "name=", "out=", "scale=", "xml", "tiles", "svg", "pdf", "flip=", "motif=", "color="])
        #print 'Opts:',opts
        #print 'Extra parameters:',extraparams
    except:
        print("")
        print("Unrecognized option:" + str(sys.argv[1:]))
        DisplayHelp()

    if len(opts) == 0:
        DisplayHelp()

    # Add additional color settings
    extra_slat_color_mapping = {}
    slat_color_mapping.update(extra_slat_color_mapping)

    directory = '.'
    scale_factor = 4
    base_file_name = None
    xml_file_name = None
    tdp_file_name = None
    tds_file_name = None
    pdf_file_name = None
    svg_file_name = None
    b_generate_tiles = False
    b_generate_xml = False
    b_generate_pdf = False
    b_generate_svg = False
    b_add_flipped_slats = True
    motif_str = None
    motif = None
    color_scheme = COLORING_BITS


    for o,p in opts:
        if o in ['-h', '--help']:
            DisplayHelp()
        elif o in ['-n', '--name']:
            base_file_name = p
        elif o in ['-o', '--out']:
            directory = p
        elif o in ['-c', '--scale']:
            scale_factor = int(p)
        elif o in ['-t', '--tiles']:
            b_generate_tiles = True
        elif o in ['-p', '--pdf']:
            b_generate_pdf = True
        elif o in ['-v', '--svg']:
            b_generate_svg = True
        elif o in ['-x', '--xml']:
            b_generate_xml = True
        elif o in ['-f', '--flip']:
            b_add_flipped_slats = True if p in ['T','t','True','true'] else False
        elif o in ['-m', '--motif']:
            motif_str = p
            motif = list(p)
        elif o in ['-c', '--color']:
            if p in ['m', 'M']:
                color_scheme = COLORING_MOTIF
            elif p not in ['b', 'B']:
                print(f"ERROR: color option must be 'b' or 'm'\n")
                DisplayHelp()
        else:
            DisplayHelp()

    if (scale_factor is None) and (not motif is None):
        scale_fact = len(motif)

    if (motif is not None) and (len(motif) != scale_factor):
        print(f'ERROR: motif len must equal the scale_factor\n')
        DisplayHelp()

    if not os.path.exists(directory):
        os.makedirs(directory)

    if base_file_name is None:
        base_file_name = f'{directory}/slat-Sierpinski-triangle-scale{scale_factor}'
        if not motif is None:
            base_file_name += f'-motif({motif_str})'

    tile_types, seed_name, gadget_slat_groups, seed_slat_groups = GetTilesAndSlats(scale_factor, motif)

    slat_system = SlatSystem()
    slat_system.b_use_rotation = False
    slat_system.binding_threshold = scale_factor

    for gadget_slat_group in gadget_slat_groups:
        for slats in [gadget_slat_group.vertical_slat_group.slats, gadget_slat_group.horizontal_slat_group.slats]:
            for slat in slats:
                if (color_scheme == COLORING_MOTIF) and (not motif is None) and (slat.name[-1] in motif):
                    SetSlatTypeToMotifColor(None, slat, motif)
                tas_slat_type = slat_system.GetPolyominoTypeByName(slat.name)
                if tas_slat_type is None:
                    tas_slat_type = GetTASSlatTypeFromLibSlatType(slat)
                    if (color_scheme == COLORING_MOTIF) and  (not motif is None) and (tas_slat_type.name[-1] in motif):
                        SetSlatTypeToMotifColor(tas_slat_type, None, motif)
                    slat_system.AddPolyominoType(tas_slat_type)
                    if b_add_flipped_slats == True:
                        rotated_slat = tas_slat_type.GetRotatedVersion(180)
                        slat_system.AddPolyominoType(rotated_slat)
    flip_text = '' if b_add_flipped_slats == False else ' (including flipped copies)'
    print(f'Slat system has {len(slat_system.PolyominoType_dict)} non-seed slat types{flip_text}')

    domain_set = GetDomainSetFromSlatSystem(slat_system)
    print(f'Slat system has {len(domain_set)} unique domains (a.k.a. handles) not including seed-specific slats:')
    #print(domain_set)

    # Get stats for potential translation/flipping errors
    horizontal_slat_dict = dict()
    vertical_slat_dict = dict()
    for slat_type_name in slat_system.PolyominoType_dict:
        slat_type = slat_system.PolyominoType_dict[slat_type_name]
        if len(slat_type.blocks) == (scale_factor * 2): #MJP: currently not checking boundary slats
            if slat_type.IsHorizontal():
                horizontal_slat_dict[slat_type.name] = slat_type
            else:
                vertical_slat_dict[slat_type.name] = slat_type
    print(f'{len(horizontal_slat_dict)} horizontal slats and {len(vertical_slat_dict)} vertical slats (before adding seed slats)')

    (half_self_check_max_overlap, self_check_max_overlap, cross_check_max_overlap,
     half_self_check_max_overlap_count, self_check_max_overlap_count, cross_check_max_overlap_count,
     max_half_self_check_example, max_self_check_example, max_cross_check_example) = GetIncorrectBindingStatsForSlats(horizontal_slat_dict, vertical_slat_dict)
    print(f'max_self_overlap: {half_self_check_max_overlap}, max_cross_overlap: {cross_check_max_overlap}')


    for seed_slat_group in seed_slat_groups:
        for slats in [seed_slat_group.vertical_slat_group.slats, seed_slat_group.horizontal_slat_group.slats]:
            for slat in slats:
                if (color_scheme == COLORING_MOTIF) and (not motif is None) and (slat.name[-1] in motif):
                    SetSlatTypeToMotifColor(None, slat, motif)
                tas_seed_slat_type = slat_system.GetPolyominoTypeByName(slat.name)
                if tas_seed_slat_type is None:
                    tas_seed_slat_type = GetTASSlatTypeFromLibSlatType(slat, b_seed_only=True)
                    if (color_scheme == COLORING_MOTIF) and (not motif is None) and (tas_seed_slat_type.name[-1] in motif):
                        SetSlatTypeToMotifColor(tas_seed_slat_type, None, motif)
                    slat_system.AddPolyominoType(tas_seed_slat_type)

                tas_slat = TASSlat(tas_seed_slat_type)
                z_coord = 1 if slat.b_horizontal else 0
                translation_coords = (slat.begin[0], slat.begin[1], z_coord)
                tas_slat.translate(translation_coords)
                slat_system.seed.AddPolyominoForViewingOnly(tas_slat)
                #print(f'added slat {tas_slat}')

    all_slat_groups = copy.deepcopy(gadget_slat_groups)
    all_slat_groups.append(seed_slat_groups[0])

    # Do Sierpinski triangle bit coloring
    if (color_scheme == COLORING_BITS):
        for slat_type_name in slat_system.PolyominoType_dict:
            slat_type = slat_system.PolyominoType_dict[slat_type_name]
            SetSierpinskiSlatColor(tas_slat_type=slat_type)

        for gadget_slat_group in all_slat_groups:
            for slats in [gadget_slat_group.vertical_slat_group.slats, gadget_slat_group.horizontal_slat_group.slats]:
                for slat in slats:
                    SetSierpinskiSlatColor(lib_slat_type=slat)
                    #print(f'slat color: {slat.color}')

    if b_generate_tiles:
        tdp_file_name = f'{base_file_name}.tdp'
        tds_file_name = f'{base_file_name}.tds'

        with open(tdp_file_name, 'w') as f:
            f.write(f'{tds_file_name}\n')
            f.write('Temperature=2\n')
            f.write(f'{seed_name} -1 -1\n')

        with open(tds_file_name, 'w') as f:
            for tile in tile_types:
                f.write(f'{tile_types[tile].Output()}\n')
        print(f"Tile system output complete to '{tds_file_name}' and '{tdp_file_name}'.")

    if b_generate_xml:
        xml_file_name = f'{base_file_name}.xml'
        slat_system.WriteToFile(xml_file_name, color_mapping=wxPython_color_mapping)
        print(f'Output slat system to {xml_file_name}')
        #seed_assembly = copy.deepcopy(seed_slat_groups)
        #OutputSlatsToXML(xml_file_name, all_slat_groups, scale_factor, seed_assembly)

    if b_generate_pdf or b_generate_svg:
        if b_generate_pdf:
            pdf_file_name = f'{base_file_name}.pdf'
        if b_generate_svg:
            svg_file_name = f'{base_file_name}.svg'
        #slat_system.seed.WriteToPDForSVG(pdfFileName=pdf_file_name, svgFileName=svg_file_name, sideLen=50)
        #Actually want to write out the gadgets!!!
        DrawSlatsToFiles(all_slat_groups, scale_factor=scale_factor, svg_file_name=svg_file_name, pdf_file_name=pdf_file_name)
